// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "Communicator.hh"
#include "ConstructArg.hh"
#include "BroadcastEvent.hh"

//
//.. constructor
//
Communicator::Communicator() : MObject (NULL)
{
  ancestor = this;
  first_channel = NULL;
  server_socket = -1;
  AddClassName();
  trash_list = new TList<MObject*>(100, 100);
}

//
//.. destructor
//
Communicator::~Communicator()
{
  while(first_channel != NULL) {
    DeleteEventChannel(first_channel);
  };
};

//
//.. SetupServer
//
void Communicator::SetupServer(unsigned short int port)
{
  if (server_socket != -1) Error("communications already started");
  port_number = port;
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(struct sockaddr_in));
  struct hostent *hp;
  char hostname[200];
  gethostname(hostname, 200);
  hp = gethostbyname(hostname);
  if (hp == NULL) Error("couldn't get hostname");
  sa.sin_family = AF_INET;
  sa.sin_port = htons(port_number);
  server_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (server_socket < 0) Error("couldn't create socket");
  if (bind(server_socket, (sockaddr*) &sa, sizeof(struct sockaddr_in)) < 0) {
    Error("couldn't bind socket");
  };
  listen(server_socket, 20);
  fcntl(server_socket, F_SETFL, O_NONBLOCK);
  host = hp->h_name;
  char new_id[1000];
  ostrstream s(new_id, 1000);
  s << hp->h_name << "[" << port_number << "]" << '\0';
  id_text = new_id;
  mesh_id = id_text;
  BroadcastEvent E("CommunicatorChanged");
  SendEvent(E);
};

//
//.. AddClientConnection
//
void Communicator::AddClientConnection(string host, unsigned short int port)
{
  struct hostent *hp;
  hp = gethostbyname(host.c_str());
  if (hp == NULL) Error("couldn't get hostname");
  string full_host = hp->h_name;
  EventChannel *channel;
  channel = new EventChannel(this, full_host, port);
  AddEventChannel(channel);
  channel->Handshake();
};

//
//.. DeleteClientConnection
//
void Communicator::DeleteClientConnection(string host, unsigned short int port)
{
  EventChannel *channel = first_channel;
  EventChannel *del_channel;
  while (channel != NULL) {
    if ((channel->remote_port == port) && (channel->remote_host == host)) {
      del_channel = channel;
      channel = channel->next_channel;
      DeleteEventChannel(del_channel);
    } else {
      channel = channel->next_channel;
    };
  };
};

//
//.. AddEventChannel
//
void Communicator::AddEventChannel(EventChannel *new_channel)
{
  EventChannel *channel;
  new_channel->next_channel = NULL;
  if (first_channel == NULL) {
    first_channel = new_channel;
    new_channel->prev_channel = NULL;
    new_channel->next_channel = NULL;
  } else {
    channel = first_channel;
    while (channel->next_channel != NULL) channel = channel->next_channel;
    channel->next_channel = new_channel;
    new_channel->prev_channel = channel;
    new_channel->next_channel = NULL;
  };
};

//
//.. DeleteEventChannel
//
void Communicator::DeleteEventChannel(EventChannel *old_channel)
{
  if (old_channel->prev_channel != NULL) {
    old_channel->prev_channel->next_channel = old_channel->next_channel;
  };
  if (old_channel->next_channel != NULL) {
    old_channel->next_channel->prev_channel = old_channel->prev_channel;
  };
  old_channel->Connect(2); // hangup!
  if (old_channel == first_channel) {
    delete old_channel;
    first_channel = NULL;
  } else {
    delete old_channel;
  };
};

//
//.. Incoming
//
void Communicator::Incoming()
{
  int sock = accept(ancestor->server_socket, NULL, NULL);
  if (sock >= 0) {
    Mesg("request accepted");
    
    // get a single character
    char c;
    read_from_sock(sock, &c, 1);
    string req_host = ReadString(sock);
    unsigned short int req_port = ReadInt(sock);

    // determine what the remote side wants
    switch(c) {
    case 0:
      // handshake
      {
	bool established = false;
	EventChannel *channel = first_channel;
	cout << "handshake from " << req_host << ":" << req_port << endl;
	while (channel != NULL) {
	  if ((channel->remote_port == req_port) && (channel->remote_host == req_host)) {
	    cout << "channel already established :-)" << endl;
	    established = true;
	    break;
	  };
	  channel = channel->next_channel;
	};
	if (!established) {
	  AddEventChannel(new EventChannel(this, req_host, req_port));
	  cout << "established channel to " << req_host << ":" << req_port << endl;
	};
	WriteString(sock, "ok");
	close(sock);
      }
    break;
    case 1:
      // event
      {
	EventChannel *channel = first_channel;
	cout << "event from " << req_host << ":" << req_port << endl;
	while (channel != NULL) {
	  if ((channel->remote_port == req_port) && (channel->remote_host == req_host)) {
	    cout << "found channel for it :-)" << endl;
	    channel->sock = sock;
	    Event E;
	    channel->Receive(E);
	    E.origin_channel = channel;
	    E.carrier = NULL;
	    ReceiveEvent(E);
	    channel->Send(E);
	    channel->Disconnect();
	    break;
	  };
	  channel = channel->next_channel;
	};
      };
    break;
    case 2:
      // hangup
      {
	EventChannel *channel = first_channel;
	EventChannel *del_channel;
	cout << "hangup from " << req_host << ":" << req_port << endl;
	while (channel != NULL) {
	  if ((channel->remote_port == req_port) && (channel->remote_host == req_host)) {
	    cout << "found channel to hangup" << endl;
	    del_channel = channel;
	    channel = channel->next_channel;
	    DeleteEventChannel(del_channel);
	  } else {
	    channel = channel->next_channel;
	  };
	};
      };
    break;
    };
  };
};
  
//
//.. HandleEvent
//
void Communicator::HandleEvent(Event &E)
{
  MObject::HandleEvent(E);
  if (!E.Done() && E.send_to_other_nodes) {
    EventChannel *channel;
    channel = first_channel;
    while (channel != NULL) {
      if (channel != E.origin_channel) {
	if (channel->Connect(1)) {
	  // connection successful for events
	  channel->Send(E);
	  channel->Receive(E);
	  channel->Disconnect();
	} else {
	  // connection unsuccessful => removing channel
	  char text[400];
	  ostrstream s(text, 400);
	  s << "lost connection to '" << channel->remote_host << ":"
	    << channel->remote_port << "'\0";
	  Mesg(text);
	  DeleteEventChannel(channel);
	};
      }
      channel = channel->next_channel;
    };
  };
};

//
//.. read_from_sock
//
void Communicator::read_from_sock(int sock, void *buff, int n)
{
  char* buf = (char*) buff;
  int bc = 0, br;
  while (bc < n) {
    br = read(sock, buf, n-bc);
    if (br < 0) Error("socket read error");
    bc += br;
    buf += br;
  };
};

//
//.. WriteInt
//
void Communicator::WriteInt(int sock, long int i)
{
  int si = htonl(i);
  write(sock, &si, 4);
}

//
//.. ReadInt
//
long int Communicator::ReadInt(int sock)
{
  long int ri;
  read_from_sock(sock, &ri, 4);
  return ntohl(ri);
}

//
//.. WriteReal
//
void Communicator::WriteReal(int sock, real r)
{
  double sr = r;
#if defined(___SYSTEM_LINUX)
  char *c;
  c = (char*) &r;
  write(sock, &c[7], 1);
  write(sock, &c[6], 1);
  write(sock, &c[5], 1);
  write(sock, &c[4], 1);
  write(sock, &c[3], 1);
  write(sock, &c[2], 1);
  write(sock, &c[1], 1);
  write(sock, &c[0], 1);
#endif
#if !defined(___SYSTEM_LINUX)
  write(sock, &sr, 8);
#endif
}

//
//.. ReadReal
//
real Communicator::ReadReal(int sock)
{
  double rr;
#if defined(___SYSTEM_LINUX)
  char *c;
  c = (char*) &rr;
  read_from_sock(sock, &c[7], 1);
  read_from_sock(sock, &c[6], 1);
  read_from_sock(sock, &c[5], 1);
  read_from_sock(sock, &c[4], 1);
  read_from_sock(sock, &c[3], 1);
  read_from_sock(sock, &c[2], 1);
  read_from_sock(sock, &c[1], 1);
  read_from_sock(sock, &c[0], 1);
#endif
#if !defined(___SYSTEM_LINUX)
  read_from_sock(sock, &rr, 8);
#endif
  return rr;
}

//
//.. WriteString
//
void Communicator::WriteString(int sock, string s)
{
  write(sock, s.c_str(), s.size() + 1);
};

//
//.. ReadString
//
string Communicator::ReadString(int sock)
{
  char c;
  string s = "";
  read_from_sock(sock, &c, 1);
  while (c != 0) {
    s = s + c;
    read_from_sock(sock, &c, 1);
  };
  return s;
};

//
//.. AddToTrash
//
void Communicator::AddToTrash(MObject *mobj)
{
  try {
    trash_list->FindItem(mobj);
  }
  catch (NotFound_error) {
    trash_list->NewEntry() = mobj;
  };
};

//
//.. DelFromTrash
//
void Communicator::DelFromTrash(MObject *mobj)
{  
  try {
    size_t i = trash_list->FindItem(mobj);
    trash_list->DelEntry(i);
  }
  catch (NotFound_error) {
    Bug();
  };
};

//
//.. CollectGarbage
//
void Communicator::CollectGarbage()
{
  FORALL(i, trash_list->) {
    delete trash_list->At(i);
  };
  delete trash_list;
  trash_list = new TList<MObject*>(100, 100);
  BroadcastEvent E("ObjectTreeChanged");
  SendEvent(E);
};







