// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik
// Universitaet Duisburg, Germany
// Institute for Combustion and Gas Dynamics
// University of Duisburg, Germany
// Sat Jun 5 1999
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!
#ifndef CompressibleBViscFlux2D_HH
#define CompressibleBViscFlux2D_HH

#include <mouse_config.h>

class CompressibleBViscFlux2D;

#include <mouse_types.hh>
#include <ConstructArg.hh>
#include <CompressibleViscFlux2D.hh>

/**
*/
class CompressibleBViscFlux2D : public CompressibleOperator2D
{
protected:
  /** the Reynolds number */ 
  real Reynolds;

  /** the reference length for the Reynolds number */
  real Ref_length;

  /** the Prandl number */
  real Prandl;

  /** the inverse reference length */
  real Re_ref;

  /** auxilary coefficient */
  real Fkenn2;

  /** auxilary coefficient */
  real Fkenn3;

  /** auxilary coefficient */
  real Fkenn4;

  /** */ VarContainer::var1_t drho_dx;
  /** */ VarContainer::var1_t drhou_dx;
  /** */ VarContainer::var1_t drhov_dx;
  /** */ VarContainer::var1_t drhoE_dx;
  /** */ VarContainer::var1_t drho_dy;
  /** */ VarContainer::var1_t drhou_dy;
  /** */ VarContainer::var1_t drhov_dy;
  /** */ VarContainer::var1_t drhoE_dy;

public:
  /** */ CompressibleBViscFlux2D (carg_t arg);
  /** */ virtual ~CompressibleBViscFlux2D ();
  /** */ void compute(size_t i, real nx, real ny,
		      real &flux_rhou, real &flux_rhov, real &flux_rhoE);
  /** */ void operate(size_t ie, size_t ip, size_t from, size_t to, 
		      real nx, real ny, bool mod_from, bool mod_to);
};

inline void CompressibleBViscFlux2D::compute(size_t i, real nx, real ny,
					     real &flux_rhou, real &flux_rhov, real &flux_rhoE)
{
  real d_u_dx, d_v_dx, d_T_dx, d_u_dy, d_v_dy, d_T_dy;
  real tau_xx, tau_xy, tau_yy, q_dot_x, q_dot_y;

  //.. Compute primitive gradients d(u)/d(.), d(v)/d(.), d(T)/d(.)
  d_u_dx = (drhou_dx[i] - drho_dx[i]*rhou[i]/rho[i])/rho[i];
  d_v_dx = (drhov_dx[i] - drho_dx[i]*rhov[i]/rho[i])/rho[i];
  d_T_dx = gamma1 / (rho[i] * rho[i]) *
    (drhoE_dx[i] * rho[i] - drho_dx[i] * rhoE[i]
     - rhou[i] * drhou_dx[i] - rhov[i] * drhov_dx[i]
     + drho_dx[i] / rho[i] * (rhou[i] * rhou[i] + rhov[i] * rhov[i]));
  d_u_dy = (drhou_dy[i] - drho_dy[i]*rhou[i]/rho[i])/rho[i];
  d_v_dy = (drhov_dy[i] - drho_dy[i]*rhov[i]/rho[i])/rho[i];
  d_T_dy = gamma1 / (rho[i] * rho[i]) *
    (drhoE_dy[i] * rho[i] - drho_dy[i] * rhoE[i]
     - rhou[i] * drhou_dy[i] - rhov[i] * drhov_dy[i]
     + drho_dy[i] / rho[i] * (rhou[i] * rhou[i] + rhov[i] * rhov[i]));
  
  //.. Compute the fluxes out of that
  //.... tensions
  tau_xx = Fkenn2 * (2. * d_u_dx       -d_v_dy);
  tau_yy = Fkenn2 * (    -d_u_dx + 2. * d_v_dy);
  tau_xy = Fkenn3 * ( d_u_dy + d_v_dx);
  //.... heat flux
  q_dot_x = Fkenn4 * d_T_dx;
  q_dot_y = Fkenn4 * d_T_dy;
  
  //.. The viscous fluxes for the cons. quantities
  flux_rhou = tau_xx * nx + tau_xy * ny;
  flux_rhov = tau_xy * nx + tau_yy * ny;
  flux_rhoE =
    ((tau_xx * rhou[i] + tau_xy * rhov[i]) / rho[i] + q_dot_x) * nx
    + ((tau_xy * rhou[i] + tau_yy * rhov[i]) / rho[i] + q_dot_y) * ny;
};  


inline void CompressibleBViscFlux2D::operate(size_t ie, size_t ip, 
					     size_t from, size_t to, 
					     real nx, real ny, 
					     bool mod_from, bool mod_to)
{
  real flux_rhou, flux_rhov, flux_rhoE;
  if (mod_to) {
    compute(to, nx, ny, flux_rhou, flux_rhov, flux_rhoE);
    res_rhou[to] -= flux_rhou;
    res_rhov[to] -= flux_rhov;
    res_rhoE[to] -= flux_rhoE;
  };
  if (mod_from) {
    compute(from, nx, ny, flux_rhou, flux_rhov, flux_rhoE);
    res_rhou[from] -= flux_rhou;
    res_rhov[from] -= flux_rhov;
    res_rhoE[from] -= flux_rhoE;
  };
};


#endif
