// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(ELEMENTLIST_HH)
#define ELEMENTLIST_HH

class ElementList;

#include "Element.hh"
#include <MeshDataList.hh>
#include <TList.hh>
#include "PointList.hh"

/**
   List of elements for a mesh.
   \\ \\
   {\bf MCL-syntax:}
   \begin{verbatim}
   ElementList { 
     <--
     string filename
     boolean read_elements
   }
   \end{verbatim}
*/
class ElementList : public MeshDataList
{
protected:
  /** the corresponding PointList. */
  PointList* PL;

  /** the element information */
  TList<Element*> *element;

public:
  /** */ ElementList (ConstructArg arg);

  /** Read Elements. Read all elements from the file specified in the control file. */
  void ReadElements ();

  /** Add a new element.
    @param n the number of nodes \\
    {\bf p_(0), p_(1), p(2) ... p(n-1)} the nodes
    @return the index of the new element */
  size_t AddElement (unsigned int n, ...);

  /** Add a new element.
    @param n the number of nodes
    @param p array with the nodes
    @return the index of the new element */
  size_t AddElement (unsigned int n, size_t *p);

  /** Get a reference to a node.
    @param e index of the element
    @param n index of the node
    @return a reference to the point-address of the node */
  size_t& Node (size_t e, unsigned int n) {
    //return *( ( (size_t*) pointer[e] ) + n);
    return element->At(e)->Node(n);
  };

  /** Get a reference to a node.
    Every value for n is allowed here.
    @param e index of the element
    @param n index of the node 
    @return a reference to the point-address of the node */
  size_t& PermNode (size_t e, int n) {
    if (n < 0) n += NumNodes(e);
    n -= (n / (NumNodes(e))) * NumNodes(e);
    return Node (e, n);
  };
  /** Get the number of nodes for an element.
    @param e the index of the element
    @return the number of nodes for this element */
  unsigned int NumNodes (size_t e) {
    //return ( (Element*) EntryType(e) ) -> num_nodes;
    return element->At(e)->NumNodes();
  };

  /** Get the middlepoint's x-coordinate of an element.
    @param e the index of the element
    @return the middlepoint's coordinate */
  real x_middle (size_t e);

  /** Get the middlepoint's y-coordinate of an element.
    @param e the index of the element
    @return the middlepoint's coordinate */
  real y_middle (size_t e);

  /** Get the middlepoint's z-coordinate of an element.
    @param e the index of the element
    @return the middlepoint's coordinate */
  real z_middle (size_t e);

  /** Write the ElementList to the file, specified in the controlfile. */
  void WriteToFile ();
};

#endif







