// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(LIST_HH)
#define LIST_HH

#include <cstddef>

class List;

#include "mouse_types.hh"

/**
   This exception will be thrown if an item cannot be found in a List,
   or one of it's derived classes.
*/
struct NotFound_error 
{
  /** */ int code;
  /** */ NotFound_error() { code = 0; };
  /** */ NotFound_error(int a_code) { code = a_code; };
};

/**
   This exception will be thrown if 'mdebug' is enabled and an index is invalid
*/
struct InvalidIndex_error
{
  int i;
  InvalidIndex_error(size_t an_i) { 
    i=i; 
  };
};

/**
   This is a dynamic list, which only provides the necessary handling
   routines and control data, but no payload. Other Lists can be linked to ensure
   a consistent length of different Lists. They will expand and shrink simultaneously.
*/
class List 
{
protected:
  /** 
      array to store the entry mapping after garbage collection 
  */
  size_t *offset;

  /**
      the maximal offset value found
  */
  size_t max_offset;

  /** 
      indicates if a certain entry is active or not 
  */
  bool *active;

  /** 
      number of client lists if this is a master 
  */
  unsigned int num_clients;

  /** 
      initialize the List 
      @param mne the initial maximum number of entries
      @param delta a delta entries to extend or shrink the list
  */
  void InitList (size_t mne, size_t delta);

  /** 
      extend or shrink the list
      @param delta the number of new entries, 
      negative values will shrink the list
  */ 
  void ExtendList(size_t delta);

  /** 
      add a new client if this is a master list
      @param client_to_add a pointer to the new client list
  */ 
  void AddClient (List *client_to_add);
  
  /**
      delete a client from this master
      @param client_to_del a pointer to the client,
      which shall be deleted
  */
  void DelClient (List *client_to_del);

  /** 
      flag to determine if the list has been initialized. 
  */
  bool initialized;

  /** 
      Initialization of an entry.
      This method does nothing for the base-class and can be overwritten in child-classes.
      @param i the index of the entry 
  */
  virtual void InitEntry (size_t i) {};

  /** 
      Extend the list.
      This method has to be overwritten in child-classes to ensure a proper 
      working of the cleaning mechanism.
      @param delta the increment 
  */
  virtual void Extend (size_t delta) {};

  /** 
      The master List, which is managing the dynamics. 
  */
  List *master;

  /** 
      a field to store pointers to the clients
  */
  List **client;

  /**
     flag to determine if the list is clean (no holes)
  */
  bool is_clean;

public:
  /** 
      the maximum number of entries 
  */
  size_t max_num_entries;

  /** 
      the current number of entries 
  */
  size_t num_entries;

  /** 
      the dynamical increment 
  */
  size_t delta_entries;

  /** 
      get the maximum number of entries 
  */
  size_t MaxNumEntries() const { return max_num_entries; };

  /** 
      get the current number of entries 
  */
  size_t NumEntries() const { return num_entries; };

  /** 
      get the dynamical increment 
  */
  size_t DeltaEntries() const { return delta_entries; };

  /**
      set the dynamical increment
      @param new_delta the new increment
  */
  void SetDelta(size_t new_delta);

  /** 
      create a new uninitialized List
  */ 
  List ();

  /** 
      create a new list as master and initialize it
      @param mne initial maximum number of entries
      @param delta initial increment of entries 
  */
  List (size_t mne, size_t delta);

  /** 
      create a new list which is linked to an existing master and
      initialize it
  */
  List (List* a_master);

  /** */ virtual ~List ();

  /** 
      get the active flag for an entry
      @param i index of the entry
      @return the active flag
  */
  bool IsActive(size_t i) const {
    return master->active[i]; 
  };

  /** 
      get the number of active entries. 
      @return the number of active entries
  */
  size_t NumActiveEntries () const;

  /** 
      add an entry at the end of the list
      @return the index of the new entry 
  */
  size_t AddEntry ();

  /** 
      delete an entry
      @param i the index of the entry to be deleted 
  */
  void DelEntry (size_t i);

  /** 
      delete all entries
  */
  void DelAll ();

  /** 
      find the next active entry. This method is a bit slow right now, but that
      will be improved in the future. So please stick to this and do not try
      to write a fast workaround.
      @param i the index of the entry whose following entry has to be found
      @return the index of the following entry 
  */
  size_t next(size_t i) const {
    do { i++; } while ((i < num_entries) && !IsActive(i));
    return i;
  };

  /** 
      find the previous active entry. This method is a bit slow right now, but that
      will be improved in the future. So please stick to this and do not try
      to write a fast workaround.
      @param i the index of the entry whose following entry has to be found
      @return the index of the following entry 
  */
  size_t prev(size_t i) const {
    if (i == 0) throw NotFound_error();
    do { i--; } while ((i > 0) && !IsActive(i));
    if (!IsActive(i)) throw NotFound_error();
    return i;
  };

  /** deprecated */
  size_t Next(size_t i) const { return next(i); };

  /** 
      find the first active entry.
      @return the index of the first active entry 
  */
  size_t begin() const;

  /**
     find the last active enty
     @return the index of the last active entry 
  */
  size_t last() const { return prev(end()); };

  /**
      specifies an index after the last entry
      @return an index after the last one
  */
  size_t end() const { return NumEntries(); };

  /** deprecated */
  size_t First() { return begin(); };

  /** 
      create the offset-list for the cleaning process
  */
  void CreateOffsetList ();

  /** 
      delete the offset list. 
  */
  void DeleteOffsetList () { 
    delete [] offset; 
    offset = NULL;
  };

  /** 
      get the offset of an entry
      @param i the index of the entry
      @return the offset of the entry 
  */
  size_t Offset (size_t i);

  /** 
      find out if the list has deleted entries or if it is clean
  */ 
  bool IsClean() const { return master->is_clean; };

  /** 
      clean the list.
      This method eliminates all holes from the list and resets num_entries.
      After CleanUp num_entries will be identical with the number of active entries. 
  */
  virtual void CleanUp ();

  /** 
      links another list to this list. 
  */
  void Link (List *linked_list);

  /** 
      copy an entry. 
      This method has to be overwritten in child-classes to ensure a proper 
      working of the cleaning mechanism.
      @param src the index of the source entry
      @param dest the index of the destination entry 
  */
  virtual void CopyEntry (size_t src, size_t dest);

  /** */ void operator=(const List &other);
};


/**
   a loop over all active entries of a List.
   This macro expands to for(size_t I = L begin(); I < L end(); I = L next(I)).
   @param I the index variable
   @param L the List, including the member-operator (for example "this->") 
*/
#define FORALL(I,L) for (size_t I = L begin(); I < L end(); I = L next(I))



#endif







