// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "Mesh2D.hh"
#include <geometry2D.hh>

//
//.. constructor
//
Mesh2D::Mesh2D(ConstructArg arg) : MObject(arg)
{
  AddClassName();
  DefineAsUniqueOnMesh();
  PL = FindObject<PointList>(false, true);
  EL = FindObject<ElementList>(false, true);
  if (EL == NULL) {
    Mesg("didn't find an ElementList, trying to construct it ...");
    EL = BuildObject<ElementList>(arg);
  }
  nd = new TList<real>(PL, 0);
  FORALL(i, PL->) nd->At(i) = 0;
  a11 = new TList<real>(PL, 1.0);
  a12 = new TList<real>(PL, 0.0);
  a22 = new TList<real>(PL, 1.0);
  //front_p1 = new TList<size_t>(100, 100);
  //front_p2 = new TList<size_t>(front_p1);
  PEL = new PelList2D(this);
  fixed = new TList<bool>(PL, false);
  marked = new TList<bool>(PL, true);
  //TRI = new Triangulator2D(this);
};

//
//.. destructor
//
Mesh2D::~Mesh2D() 
{
  //delete front_p1;
  //delete front_p2;
};

//
//.. UpdatePEL
//
void Mesh2D::UpdatePEL () 
{
  if (PEL == NULL) PEL = new PelList2D (this);
  PEL->Create ();
};

//
//.. CheckOrientation
//
bool Mesh2D::CheckOrientation (size_t i_element) {
  bool ok = true;
  size_t i_node, p1, p2, p3;
  double alpha;
  // compute sum of all inner angles
  alpha = 0;
  size_t nn = EL->NumNodes(i_element);
  for (i_node = 0; i_node < nn; i_node++) {
    p1 = EL->Node (i_element, i_node);
    if (i_node + 1 >= nn) {
      p2 = EL->Node (i_element, i_node + 1 - nn);
    } else {
      p2 = EL->Node (i_element, i_node + 1);
    };
    if (i_node + 2 >= nn) {
      p3 = EL->Node (i_element, i_node + 2 - nn);
    } else {
      p3 = EL->Node (i_element, i_node + 2);
    };
    alpha += Angle (PL->x(p2) - PL->x(p1), PL->y(p2) - PL->y(p1), 
		    PL->x(p3) - PL->x(p2), PL->y(p3) - PL->y(p2));
  };
  // change orientation if necessary
  if (alpha < 0) {
    ok = false;
    size_t *node_h;
    node_h = new size_t [nn];
    for (i_node = 0; i_node < nn; i_node++) {
      node_h[i_node] = EL->Node (i_element, nn - 1 - i_node);
    };
    for (i_node = 0; i_node < nn; i_node++) {
      EL->Node (i_element, i_node) = node_h[i_node];
    };
    delete [] node_h;
  };
  return ok;
};

//
//.. WriteMeshData
//
void Mesh2D::WriteMeshData () {
  PL->WriteAllPoints();
  EL->WriteToFile();
};

//
//.. Transform
//
void Mesh2D::Transform(size_t p0, real &x, real &y)
{
  Transform(PL->x(p0), PL->y(p0), A11(p0), A12(p0), A22(p0), x, y);
}

void Mesh2D::Transform(real x0, real y0, 
		       real a11, real a12, real a22,
		       real &x, real &y)
{
  x -= x0;
  y -= y0;
  real x_n = x0 + a11 * x + a12 * y;
  real y_n = y0 + a12 * x + a22 * y;
  x = x_n;
  y = y_n;
};

//
//.. Retransform
//
void Mesh2D::Retransform(size_t p0, real &x, real &y)
{
  Retransform(PL->x(p0), PL->y(p0), A11(p0), A12(p0), A22(p0), x, y);
};

void Mesh2D::Retransform(real x0, real y0, 
			 real a11, real a12, real a22,
			 real &x, real &y)
{
  real det = (a11*a22-a12*a12);
  real a11_i = a22 / det;
  real a12_i = -a12 / det;
  real a22_i = a11 / det;
  x -= x0;
  y -= y0;
  real x_n = x0 + a11_i * x + a12_i * y;
  real y_n = y0 + a12_i * x + a22_i * y;
  x = x_n;
  y = y_n;
};

void Mesh2D::CalcEigenvalues(real a11, real a12, real a22, real &l1, real &l2)
{
  l1 = a11/2+a22/2+sqrt(a11*a11-2.0*a11*a22+a22*a22+4.0*a12*a12)/2;
  l2 = a11/2+a22/2-sqrt(a11*a11-2.0*a11*a22+a22*a22+4.0*a12*a12)/2;
  /*
  real sr = sqrt(a12*a12 + .25*(a11*a11 + a22*a22) -.5*a11*a22);
  l1 = .5*(a11+a22) + sr;
  l2 = .5*(a11+a22) - sr;
  */
  if (l1 < l2) {
    real h = l1;
    l1 = l2;
    l2 = h;
  };
};

//
//.. CalcDirVector
//
void Mesh2D::CalcDirVector(size_t p, real &vx, real &vy)
{
  size_t i, n, pa, pb;
  real nx, ny, d, l1, l2, a11, a12, a22;
  vx = 0;
  vy = 0;
  for (i = 0; i < PEL->Size(p); i++) {
    n = FindNode(PEL->At(p, i), p);
    pa = EL->PermNode(PEL->At(p, i), n + 1);
    pb = EL->PermNode(PEL->At(p, i), n - 1);
    nx = PL->y(pb) - PL->y(pa);
    ny = PL->x(pa) - PL->x(pb);
    a11 = .5*(A11(pa) + A11(pb));
    a12 = .5*(A12(pa) + A12(pb));
    a22 = .5*(A22(pa) + A22(pb));
    CalcEigenvalues(a11, a12, a22, l1, l2);
    if (PL->code(p) == 0) {
      vx -= nx*l1/l2;
      vy -= ny*l1/l2;
    } else {
      vx += nx;
      vy += ny;
    };
  };
  d = DIST(0, 0, vx, vy);
  if (d < 1e-4) {
    vx = 0;
    vy = 1;
  } else {
    vx /= d;
    vy /= d;
  };
};

//
//.. AddFront
//
/*
void Mesh2D::AddFront(size_t p1, size_t p2)
{
  size_t i;
  if ((PL->code(p1) != 0) && (PL->code(p2) != 0)) {
    i = front_p1->AddEntry();
    front_p1->At(i) = p1;
    front_p2->At(i) = p2;
  };
  real dx = PL->x(p2) - PL->x(p1);
  real dy = PL->y(p2) - PL->y(p1);
  nd->At(p1) += .5/DIST(0, 0, dx, dy);
  nd->At(p2) += .5/DIST(0, 0, dx, dy);
  a11->At(p1) = 1;
  a12->At(p1) = 0;
  a22->At(p1) = 1;
  a11->At(p2) = 1;
  a12->At(p2) = 0;
  a22->At(p2) = 1;
};
*/

//
//.. SetMatrix
//
void Mesh2D::SetMatrix(size_t point, real R1, real R2, real gamma)
{
  a11->At(point) = R2 - R2 * SQR(cos(gamma)) + R1 * SQR(cos(gamma));
  a12->At(point) = R1 * sin(gamma) * cos(gamma) - R2 * sin(gamma) * cos(gamma);
  a22->At(point) = R1 - R1 * SQR(cos(gamma)) + R2 * SQR(cos(gamma));
};

//
//.. FindNode
//
size_t Mesh2D::FindNode(size_t element, size_t point)
{
  for (size_t i = 0; i <= EL->NumNodes(element); i++) {
    if (EL->Node(element, i) == point) return i;
  }
  throw NotFound_error();
};

//
//.. FindElements
//
void Mesh2D::FindElements (size_t p1, size_t p2, 
			   size_t &e1, size_t &e2,
			   bool &f1, bool &f2) 
{
  size_t h, e[2] = {0, 0};
  bool fh, f[2] = {false, false};
  size_t i, j, n=0;
  // find the entries
  for (i = 0; i < PEL->Size(p1); i++) {
    for (j = 0; j < PEL->Size(p2); j++) {
      if (PEL->At(p1, i) == PEL->At(p2, j)) {
	e[n] = PEL->At(p1, i);
	f[n] = true;
	n++;
	break;
      };
    };
    if (n == 2) break;
  };
  // check orientation
  if (f[0]) {
    n = FindNode (e[0], p1);
    if (EL->PermNode(e[0], n+1) != p2) {
      h = e[0];
      e[0] = e[1];
      e[1] = h;
      fh = f[0];
      f[0] = f[1];
      f[1] = fh;
    };
  };
  e1 = e[0];
  e2 = e[1];
  f1 = f[0];
  f2 = f[1];
};

//
//.. CleanUp
//
void Mesh2D::CleanUp () {
  PL->CleanUp();
  size_t i_node;
  size_t nn;
  EL->CleanUp();
  FORALL(i_element, EL->) {
    nn = EL->NumNodes(i_element);
    for (i_node = 0; i_node < nn; i_node++) {
      EL->Node(i_element, i_node) -= PL->Offset(EL->Node(i_element, i_node));
    };
  };
  /*
  FORALL(i_front, front_p1->) {
    front_p1->At(i_front) -= PL->Offset(front_p1->At(i_front));
    front_p2->At(i_front) -= PL->Offset(front_p2->At(i_front));
  };
  */
  PEL->Create();
};

//
//.. CheckPoint
//.... check if an edge can be eliminated
//
bool Mesh2D::CheckPoint (size_t p1, size_t p2)
{
  int n;
  size_t e1, e2, pl, pr;
  size_t dummy, ee1, ee2, ee3, ee4;
  bool fe1, fe2, fdummy, fee1, fee2, fee3, fee4;
  FindElements(p1, p2, e1, e2, fe1, fe2);
  if (PL->code(p1) != 0) return false;
  if (!fe1 || !fe2) return false;
  if ((EL->NumNodes(e1) != 3) || (EL->NumNodes(e2) != 3)) return false;
  n = FindNode(e1, p2);
  pl = EL->PermNode(e1, n+1);
  n = FindNode(e2, p1);
  pr = EL->PermNode(e2, n+1);
  FindElements(pl, p1, dummy, ee1, fdummy, fee1);
  FindElements(p1, pr, dummy, ee2, fdummy, fee2);
  FindElements(pr, p2, dummy, ee3, fdummy, fee3);
  FindElements(p2, pl, dummy, ee4, fdummy, fee4);
  if (fee1 && fee4) if (ee1 == ee4) return false;
  if (fee2 && fee3) if (ee2 == ee3) return false;
  if ((PL->code(pl) == 0) && (PEL->Size(pl) < 4)) return false;
  if ((PL->code(pr) == 0) && (PEL->Size(pr) < 4)) return false;
  return CheckPoint(p1, PL->x(p2), PL->y(p2), e1, e2);
};

//
//.. CheckPoint
//
bool Mesh2D::CheckPoint (size_t point, real x_new, real y_new, int ne,
			    size_t excl_elem1, size_t excl_elem2) 
{
  if ((PL->code(point) != 0) || fixed->At(point)) return false;
  real x_old = PL->x(point);
  real y_old = PL->y(point);
  PL->x(point) = x_new;
  PL->y(point) = y_new;
  bool ok = true;
  bool check_elem;
  size_t element;
  int j;
  real vx[4], vy[4];
  size_t pn1, pn2;
  for (size_t i = 0; i < PEL->Size(point); i++) {
    element = PEL->At(point, i);
    check_elem = true;
    if (ne > 0) check_elem = check_elem && (element != excl_elem1);
    if (ne > 1) check_elem = check_elem && (element != excl_elem2);
    //if ((element != excl_elem1) && (element != excl_elem2)) {
    if (check_elem) {
      // check if area is >= 0
      for (j = 0; j < 4; j++) {
	pn1 = EL->PermNode(element, j);
	pn2 = EL->PermNode(element, j+1);
	vx[j] = PL->x(pn2) - PL->x(pn1);
	vy[j] = PL->y(pn2) - PL->y(pn1);
      };
      for (j = 0; j < 3; j++) {
// 	if (!(vx[j]*vy[j+1] - vy[j]*vx[j+1] > 0)) {
	//cout << "alpha = " << Angle(vx[j], vy[j], vx[j+1], vy[j+1]) << endl;
 	if (Angle(vx[j], vy[j], vx[j+1], vy[j+1]) < 0) {
	  ok = false;
	  break;
	};
      };
      if (!ok) break;
    };
  };
  if (PL->code(point) != 0) ok = false;
  PL->x(point) = x_old;
  PL->y(point) = y_old;
  return ok;
};

//
//.. SetMatrix
//
void Mesh2D::SetMatrixComp(size_t i, real a_11, real a_12, real a_22)
{
  a11->At(i) = a_11;
  a12->At(i) = a_12;
  a22->At(i) = a_22;
};

//
//.. UnionPoints
//.... updating of PEL is uncertain @@
//
void Mesh2D::UnionPoints (size_t p, size_t p_pullto) 
{
  size_t i, j;
  size_t e1, e2, e;
  bool f1, f2;
  FindElements(p, p_pullto, e1, e2, f1, f2);
  if (EL->NumNodes(e1) == 4) Error("trying to delete a quad directly!");
  if (EL->NumNodes(e2) == 4) Error("trying to delete a quad directly!");
  if (!CheckOrientation(e1)) {
    cout << e1 << endl;
    Error("found element with wrong orientation");
  };
  if (!CheckOrientation(e2)) {
    cout << e2 << endl;
    Error("found element with wrong orientation");
  };
  // loop over all neighbouring elements of point to be deleted
  for (i = 0; i < PEL->Size(p); i++) {
    e = PEL->At(p, i);
    if ((e == e1) || (e == e2)) {
      PEL->DeleteFrom(e);
      EL->DelEntry(e);
      i--;
    } else {
      // replace the deleted point with the point where the deleted point will be pulled to
      // in this element
      for (j = 0; j < EL->NumNodes(e); j++) {
	if (EL->Node(e, j) == p) {
	  EL->Node (e, j) = p_pullto;
	};
      };
      if (!PEL->IsIn(p_pullto, e)) PEL->AddTo(p_pullto, e);
    };
  };
  PL->DelEntry (p);
  for (i = 0; i < PEL->Size(p_pullto); i++) {
    e = PEL->At(p_pullto, i);
    if (!CheckOrientation(e)) {
      cout << EL->NumNodes(e) << endl;
      Error("element with wrong orientation created during union of points");
    };
  };
};

//
//.. FindOutsideElement
//
size_t Mesh2D::FindOutsideElement(size_t p)
{
  size_t oe;
  bool found = false;
  real s_max = -1e6;
  real vx, vy, dx, dy, s;
  CalcDirVector(p, vx, vy);
  for (size_t i = 0; i < PEL->Size(p); i++) {
    dx = EL->x_middle(PEL->At(p, i)) - PL->x(p);
    dy = EL->y_middle(PEL->At(p, i)) - PL->y(p);
    s = vx*dx + vy*dy;
    if (s > s_max) {
      oe = PEL->At(p, i);
      found = true;
      s_max = s;
    };
  };
  if (found) return oe;
  else throw NotFound_error();
};

//
//.. NumTriangles
//
int Mesh2D::NumTriangles(size_t p)
{
  int nt = 0;
  for (size_t i = 0; i < PEL->Size(p); i++) {
    if (EL->NumNodes(PEL->At(p, i)) == 3) nt++;
  };
  return nt;
};

//
//.. NumQuads
//
int Mesh2D::NumQuads(size_t p)
{
  int nq = 0;
  for (size_t i = 0; i < PEL->Size(p); i++) {
    if (EL->NumNodes(PEL->At(p, i)) == 4) nq++;
  };
  return nq;
};
			    
//
//.. CalcBaseMatrix
//
void Mesh2D::CalcBaseMatrix (real &g11, real &g12, real &g21, real &g22,
			     real a11, real a12, real a22)
{
  real t1 = a11*a11;
  real t3 = a22*a22;
  real t6 = sqrt(t1-2.0*a11*a22+t3+4.0*a12*a12);
  g11 = a11/2-a22/2+t6/2;
  g12 = a12;
  g21 = a12;
  g22 = a22/2-a11/2-t6/2;
};

void Mesh2D::EigenProperties (real a11, real a12, real a22,
			      real &lb, real &ls, real &ebx, real &eby, real &esx, real &esy)
{
  CalcEigenvalues(a11, a12, a22, lb, ls);
  ebx = 1;
  eby = 0;
  esx = 0;
  esy = 1;
  if (lb/ls > 1.1) {
    if (fabs(lb-a22) > fabs(lb-a11)) {
      ebx = lb-a22;
      eby = a12;
    } else {
      ebx = a12;
      eby = lb-a11;
    };
    if (fabs(ls-a22) > fabs(ls-a11)) {
      esx = ls-a22;
      esy = a12;
    } else {
      esx = a12;
      esy = ls-a11;
    };
  };
  if (Angle(esx, esy, ebx, eby) < 0) {
    ebx *= -1;
    eby *= -1;
  };
  real abs_b = 1.0/DIST(0,0,ebx,eby);
  real abs_s = 1.0/DIST(0,0,esx,esy);
  ebx *= abs_b;
  eby *= abs_b;
  esx *= abs_s;
  esy *= abs_s;
};

//
//.. AreNeighbours
//
bool Mesh2D::AreNeighbours(size_t elem1, size_t elem2)
{
  size_t p1, p2, e1, e2;
  bool f1, f2;
  for (size_t node = 0; node < EL->NumNodes(elem1) - 1; node++) {
    p1 = EL->Node(elem1, node);
    p2 = EL->Node(elem1, node + 1);
    FindElements(p1, p2, e1, e2, f1, f2);
    if ((e2 == elem2) && f1 && f2) return true;
  };
  return false;
};

//
//.. MarkAll
//
void Mesh2D::MarkAll()
{
  FORALL(i, marked->) marked->At(i) = true;
};

//
//.. UnmarkAll
//
void Mesh2D::UnmarkAll()
{
  FORALL(i, marked->) marked->At(i) = false;
};

//
//.. ElementMarked
//
bool Mesh2D::ElementMarked(size_t e)
{
  for (size_t n = 0; n < EL->NumNodes(e); n++) if (marked->At(EL->Node(e, n))) return true;
  return false;
};

//
//.. MarkPoints
//
void Mesh2D::MarkPoints(TList<size_t> *point, int depth)
{
  size_t j, n;
  size_t p, e;
  if (depth == 0) {
    UnmarkAll();
    FORALL(i, point->) marked->At(point->At(i)) = true;
  } else {
    TList<size_t> new_point(100, 100);
    FORALL(i, point->) marked->At(point->At(i)) = true;
    FORALL(i, point->) {
      p = point->At(i);
      for (j = 0; j < PEL->Size(p); j++) {
	e = PEL->At(p, j);
	for (n = 0; n < EL->NumNodes(e); n++) {
	  if (!marked->At(EL->Node(e, n))) {
	    new_point.NewEntry() = EL->Node(e, n);
	    marked->At(EL->Node(e, n)) = true;
	  };
	};
      };
    };
    MarkPoints(&new_point, depth - 1);
    FORALL(i, point->) marked->At(point->At(i)) = true;
  };
};

//
//.. DeleteElement
//
void Mesh2D::DeleteElement(size_t e)
{
  PEL->DeleteFrom(e);
  EL->DelEntry(e);
};

//
//.. AddTri
//
size_t Mesh2D::AddTri(size_t p0, size_t p1, size_t p2)
{
  size_t e = EL->AddElement(3, p0, p1, p2);
  CheckOrientation(e);
  PEL->AddTo(p0, e);
  PEL->AddTo(p1, e);
  PEL->AddTo(p2, e);
  return e;
};

//
//.. AddQuad
//
size_t Mesh2D::AddQuad(size_t p0, size_t p1, size_t p2, size_t p3)
{
  size_t e = EL->AddElement(4, p0, p1, p2, p3);
  PEL->AddTo(p0, e);
  PEL->AddTo(p1, e);
  PEL->AddTo(p2, e);
  PEL->AddTo(p3, e);
  return e;
};









