// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "PelList2D.hh"
#include <ConstructArg.hh>
#include <geometry2D.hh>

//
//.. constructor @@
//
PelList2D::PelList2D (ConstructArg arg) : MObject(arg)
{
  AddClassName();
  DefineAsUniqueOnMesh();
  PL = FindObject<PointList>(false, true);
  EL = FindObject<ElementList>(false, true);
  PEL = new Flex2DArray<size_t>(PL);
  Create();
};
  
//
//.. IsIn
//
bool PelList2D::IsIn (size_t point, size_t element) 
{
  for (size_t i = 0; i < Size(point); i++) {
    if (At(point, i) == element) {
      return true;
      break;
    };
  };
  return false;
};

//
//.. Create
//
void PelList2D::Create () 
{
  size_t nn, i_node, point;
  FORALL (i, PEL->) PEL->DelSubList(i);
  FORALL (i_element, EL->) {
    nn = EL->NumNodes (i_element);
    for (i_node = 0; i_node < nn; i_node++) {
      point = EL->Node(i_element, i_node);
      if (!IsIn(point, i_element)) AddTo (point, i_element);
    };      
  };
};

//
//.. AddTo
//
void PelList2D::AddTo (size_t point, size_t element) 
{
  real alpha = Angle (1, 0, EL->x_middle(element) - PL->x(point), 
			EL->y_middle(element) - PL->y(point));
  size_t i = 0;
  while ((i < Size(point)) 
	 && (alpha > Angle (1, 0, EL->x_middle(At (point, i)) - PL->x(point),
			    EL->y_middle(At (point, i)) - PL->y(point)))) i++;
  PEL->NewAt (point, i) = element;
};
  
//
//.. DeleteFrom
//
void PelList2D::DeleteFrom (size_t element) 
{
  size_t point, i, j;
  size_t nn = EL->NumNodes(element);
  for (i = 0; i < nn; i++) {
    point = EL->Node(element, i);
    for (j = 0; j < Size (point); j++) {
      if (At (point, j) == element) {
	PEL->DelItem(point, j);
	j--;
      };
    };
  };
};

//
//.. LocalBorder
//
TList<size_t>* PelList2D::LocalBorder (size_t point) 
{
  size_t element;
  size_t i, j, k;
  TList<size_t> *local_border;
  local_border = new TList<size_t> (10, 5);
  for (i = 0; i < Size(point); i++) {
    j = 0;
    element = At(point, i);
    while (EL->Node(element, j) != point) j++;
    j++;
    while (EL->PermNode(element, j + 1) != point) {
      k = local_border->AddEntry();
      local_border->At (k) = EL->PermNode (element, j);
      j++;
    };
    if (i == Size(point)-1) {
      k = local_border->AddEntry();
      local_border->At (k) = EL->PermNode (element, j);
    };
  };
  k = local_border->num_entries - 1;
  if (local_border->At(k) != local_border->At(0)) {
    k = local_border->AddEntry();
    local_border->At(k) = local_border->At(0);
  };
  return local_border;
};

//
//.. LocalInnerBorder  @@ make same changings as to LocalBorder soon !!!!!
//
TList<size_t>* PelList2D::LocalInnerBorder (size_t point) 
{
  size_t element;
  size_t i, j, k;
  TList<size_t> *local_border;
  local_border = new TList<size_t> (10, 5);
  for (i = 0; i < Size(point); i++) {
    j = 0;
    element = At(point, i);
    while (EL->PermNode(element, j) != point) j++;
    j++;
    k = local_border->AddEntry();
    local_border->At (k) = EL->PermNode (element, j);
  };
  k = local_border->AddEntry();
  local_border->At(k) = local_border->At(0);
  return local_border;
};

//
//.. LocalPOP
//
TList<size_t>* PelList2D::LocalPOP (size_t point) 
{
  size_t element;
  size_t i, j, k;
  TList<size_t> *local_POP;
  local_POP = new TList<size_t> (10, 5);
  for (i = 0; i < Size(point); i++) {
    j = 0;
    element = At(point, i);
    while (EL->PermNode(element, j) != point) j++;
    k = local_POP->AddEntry();
    local_POP->At (k) = EL->PermNode (element, j+1);
    if (PL->code(EL->PermNode (element, j-1)) != 0) {
      k = local_POP->AddEntry();
      local_POP->At (k) = EL->PermNode (element, j-1);
    }
  };
  return local_POP;
};

//
//.. QuadConnected
//
int PelList2D::QuadConnected (size_t point) 
{
  size_t i, qc = 0;
  for (i = 0; i < Size(point); i++) if (EL->NumNodes(At(point, i)) == 4) qc++;
  return qc;
};

//
//.. TriConnected
//
int PelList2D::TriConnected (size_t point) 
{
  size_t i, tc = 0;
  for (i = 0; i < Size(point); i++) if (EL->NumNodes(At(point, i)) == 3) tc++;
  return tc;
};







