// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "PointList.hh"
#include "BroadcastEvent.hh"

//
//.. constructor
//
PointList::PointList (ConstructArg arg) : MeshDataList (arg)
{
  AddClassName();
  DefineAsUniqueOnMesh();
  filename = filename + ".points";

  // initialize the data fields
  x_point = new TList<real>(this);
  y_point = new TList<real>(this);
  if (dimension == 3) z_point = new TList<real>(this);
  else z_point = NULL;
  code_point = new TList<int>(this, 0);

  // read points if wanted
  bool psw = arg.P->ReadBool();
  if (psw) ReadAllPoints();

  ch_dist = new TList<real>(this, 0);
  ivol = new TList<real>(this, 0);
  volset = new TList<bool>(this, false);
};

//
//.. ~PointList
//
PointList::~PointList () 
{
  delete x_point;
  delete y_point;
  delete z_point;
  delete code_point;
};

//
//.. CopyEntry
//
void PointList::CopyEntry (size_t source_point, size_t dest_point) 
{
  MeshDataList::CopyEntry(source_point, dest_point);
  //@@ .... ???
};

//
//.. ReadPoint
//
void PointList::ReadPoint (size_t i, istream &s) 
{
  s >> x(i) >> y(i);
  if (dimension == 3) s >> z(i);
  s >> code(i);
};

//
//.. WritePoint
//
void PointList::WritePoint (size_t i, ostream &s) 
{
  s << x(i) << "  " << y(i);
  if (dimension == 3) s << z(i);
  s << "    " << code(i) << "\n";
};

//
//.. WriteAllPoints
//
void PointList::WriteAllPoints () 
{
  ofstream s(filename.c_str());
  s << NumActiveEntries() << "\n";
  for(size_t i_point=begin(); i_point<end(); i_point = next(i_point)) {
    WritePoint (i_point, s);
  };
};

//
//.. ReadAllPoints
//
void PointList::ReadAllPoints () 
{
  long int size, entry;
  ifstream s(filename.c_str());
  if (s.fail()) Error("couldn't open '" + filename + "'");
  s >> size;
  for (long int i = 0; i < size; i++) {
    entry = AddEntry();
    ReadPoint (entry, s);
  };
};

//
//.. FindUndefBC @@
//
bool PointList::FindUndefBC(bool gun) {
  bool undef_flag = false;

  //.. Get a List of Errors
  //
  TList<int> *undef_codes;
  undef_codes = new TList<int> (10, 5);

  //.. Spool all points and search non defined BC-codes
  //   (this is the case if a code is != 0, but not found in BC-Lists)
  //
  for(size_t k_point = begin(); k_point < end(); k_point = next(k_point)) {
    if(code(k_point) != 0) {
      if(!IsBoundary(k_point)) {
	//.... This is an undefined code
	try {
	  undef_codes->FindItem(code(k_point));
	}
	catch (NotFound_error) {
	  undef_codes->NewEntry() = code(k_point);
	  undef_flag = true;
	};
      };
    };
  };

  //.. Write warning message
  if(undef_flag) {
    cout << " WARNING, PointList::FindUndefBC(), "
	 << "following BC-codes are undefined:" << endl;
    FORALL(k_undef, undef_codes->) {
      cout << "   " << undef_codes->At(k_undef);
    };
    cout << endl;
  };

  //.. Stop execution on error and if requested
  if(gun && undef_flag) {
    cout << " Stop on warning " << endl;
    exit(EXIT_FAILURE);
  } else if(undef_flag) {
    cout << " .. continuing .." << endl;
    return true;
  };
  return false;
};

//
//.. Extend
//
void PointList::Extend (size_t delta) {
  MeshDataList::Extend(delta);
  BroadcastEvent E("Changed.PointList");
};
















