// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "ReadPicFile.hh"
#include <strstream.h>
#include <geometry2D.hh>

ReadPicFile::Line::Line(int a_code)
{
  x = new TList<real>(10, 10);
  y = new TList<real>(x);
  min_dist = -1;
  std_code = a_code;
  code = new TList<int>(x);
};

void ReadPicFile::Line::Add(real a_x, real a_y, int a_code)
{
  size_t i = x->AddEntry();
  x->At(i) = a_x;
  y->At(i) = a_y;
  if (a_code == 0) a_code = std_code;
  code->At(i) = a_code;
};

void ReadPicFile::Line::ComputeMin()
{
  size_t i = x->begin();
  size_t j = x->next(i);
  min_dist = DIST((*x)[i], (*y)[i], (*x)[j], (*y)[j]);
  i = j;
  j = x->next(i);
  while (j < x->end()) {
    if (DIST((*x)[i], (*y)[i], (*x)[j], (*y)[j]) < min_dist) {
      min_dist = DIST((*x)[i], (*y)[i], (*x)[j], (*y)[j]);
    };
    i = j;
    j = x->next(i);
  };
};

void ReadPicFile::Line::Merge(Line *line, int new_code)
{
  (*code)[code->NumEntries()-1] = new_code;
  bool first = true;
  FORALL(i, line->x->) {
    if (first) {
      first = false;
    } else {
      Add(line->X(i), line->Y(i), line->std_code);
    };
  };
};

ReadPicFile::ReadPicFile(carg_t arg) : Mesh2DGeometryAction(arg)
{
  AddClassName();
  filename = arg.P->ReadString();
  filename += ".pic";
  new_code = 1;
};

void ReadPicFile::Process(string text_line, ReadPicFile::Line* &line)
{
  line = new Line(new_code++);
  for(string::iterator i = text_line.begin(); i != text_line.end(); i++) {
    if (*i == ',') *i = ' ';
  };
  istrstream f(text_line.c_str());
  if (!f.eof()) {
    string word;
    f >> word;
    cout << "word='" << word << "'" << endl;
    if (word == "line") {
      while (!f.eof()) {
	f >> word;
	if (!f.fail()) {
	  real x, y;
	  f >> x >> y;
	  cout << x << " " << y << endl;
	  line->Add(x, y);
	};
      };
    };
  };
  line->ComputeMin();
  //return line;
};

void ReadPicFile::Act()
{
  TList<Line*> *line = new TList<Line*>(5,5,NULL);
  ifstream f(filename.c_str());
  while (!f.eof()) {
    string text_line;
    char c;
    do {
      f.get(c);
      text_line += c;
    } while (c != '\n');
    istrstream ls(text_line.c_str());
    string word;
    ls >> word;
    if (word == "line") {
      Process(text_line, line->NewEntry());
    };
  };
  bool first = true;
  real min_dist = 0;
  FORALL(i, line->) {
    if (first) {
      first = false;
      min_dist = (*line)[i]->min_dist;
    } else {
      if ((*line)[i]->min_dist < min_dist) min_dist = (*line)[i]->min_dist;
    };
  };

  // check for lines to merge
  FORALL(i, line->) {
    FORALL(j, line->) {
      if (i != j) {
	real x1 = line->At(i)->LastX();
	real y1 = line->At(i)->LastY();
	real x2 = line->At(j)->FirstX();
	real y2 = line->At(j)->FirstY();
	real d = DIST(x1, y1, x2, y2);
	if (d < .1*min_dist) {
	  line->At(i)->Merge(line->At(j), new_code++);
	  line->DelEntry(j);
	};
      };
    };
  };

  // create edges
  FORALL(i, line->) {
    bool first = true;
    size_t p0, p1, p2;
    size_t count = 0;
    FORALL(j, line->At(i)->x->) {
      if (first) {
	first = false;
	p1 = PL->AddEntry();
	p0 = p1;
	MESH->FixPoint(p1);
	MESH->SetMatrixComp(p1, 0, 0, 0);
	MESH->SetNetDensity(p1, 0);
	PL->x(p1) = line->At(i)->X(j);
	PL->y(p1) = line->At(i)->Y(j);
	PL->code(p1) = new_code++;
	count++;
      } else {
	if (count < line->At(i)->x->NumEntries()-1) {
	  p2 = PL->AddEntry();
	  MESH->FixPoint(p2);
	  MESH->SetMatrixComp(p2, 0, 0, 0);
	  MESH->SetNetDensity(p2, 0);
	  PL->x(p2) = line->At(i)->X(j);
	  PL->y(p2) = line->At(i)->Y(j);
	  PL->code(p2) = line->At(i)->code->At(j);
	  TRI->AddFront(p1, p2);
	  count++;
	  p1 = p2;
	};
      };
    };
    TRI->AddFront(p2, p0);
  };
};










