// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#ifndef Roe2D_HH
#define Roe2D_HH

class Roe2D;

#include <CompressibleFFlux2D.hh>
#include <TList.hh>
#include <MStringList.hh>

/**
   an implementation of the Roe splitting for compressible gas flow,
   including the convection of N species. 
   Thanks to Imad Elmahi for writing this flux for MOUSE.
*/
class Roe2D : public CompressibleFFlux2D
{
protected:
//   /** */ MStringList *spec_name;
//   /** */ TList<VarContainer::var1_t> *res_rhosp;
//   /** */ TList<VarContainer::var1_t> *rhosp;
//   /** */ TList<VarContainer::var1_t> *rhosp_av;
//   /** */ TList<VarContainer::var1_t> *rhosp_from;
//   /** */ TList<VarContainer::var1_t> *rhosp_to;
//   /** */ uint num_spec;
//   /** */ real *flux_rhosp;
//   /** */ real *ysp_ave;

public:
  /** */ Roe2D(carg_t arg);
  /** */ virtual ~Roe2D();
  /** */ static void compute(real rho_f, real rhou_f, real rhov_f, real rhoE_f,
			     real rho_t, real rhou_t, real rhov_t, real rhoE_t,
			     real nx, real ny, real gam,
			     real &flux_rho, real &flux_rhou, 
			     real &flux_rhov, real &flux_rhoE);
  /** */ void operate(size_t ie, size_t ip, size_t from, size_t to, 
		      real nx, real ny, bool mod_from, bool mod_to);
};

inline void Roe2D::compute(real rho_f, real rhou_f, real rhov_f, real rhoE_f,
			   real rho_t, real rhou_t, real rhov_t, real rhoE_t,
			   real nx, real ny, real gam,
			   real &flux_rho, real &flux_rhou, 
			   real &flux_rhov, real &flux_rhoE)
{
  real alpha1,alpha2,alpha3,alpha4,a1,a2,a3,a4,b1,b2,b3,b4,alpha;
  real ep_t, ep_f, p_t, p_f;
  real rhouv_t, rhouv_f;
  real lambda_f, lambda_t;
  real delta_rho, delta_rhou, delta_rhov, delta_rhoE;
  real hnx, hny;
  real tempo11, usc;

  flux_rho = 0;
  flux_rhou = 0;
  flux_rhov = 0;
  flux_rhoE = 0;

  real gam1 = gam - 1.0;
  real gam2 = gam * (gam - 1.0);
  real igam = 1.0 / (gam - 1.0);

  //
  //.. Compute the ROE Averages
  //
  //.... some useful quantities
  real coef1 = sqrt(rho_f);
  real coef2 = sqrt(rho_t);
  real somme_coef = coef1 + coef2;
  real u_f = rhou_f/rho_f;
  real v_f = rhov_f/rho_f;
  real h_f = (gam * rhoE_f/rho_f) - (gam1 / 2.0)
    * (u_f * u_f + v_f * v_f);
  real u_t = rhou_t/rho_t;
  real v_t = rhov_t/rho_t;
  real h_t = (gam * rhoE_t/rho_t) - (gam1 / 2.0)
    * (u_t * u_t + v_t * v_t);
  
  //.... averages
  //real rho_ave = coef1 * coef2;
  real u_ave = (coef1 * u_f + coef2 * u_t) / somme_coef;
  real v_ave = (coef1 * v_f + coef2 * v_t) / somme_coef;
  real h_ave = (coef1 * h_f + coef2 * h_t) / somme_coef;  
  //
  //.. Compute Speed of sound
  real scal = u_ave * nx + v_ave * ny;
  real norme = sqrt(nx * nx + ny * ny);
  real u2pv2 = u_ave * u_ave + v_ave * v_ave;
  real c_speed = gam1 * (h_ave - 0.5 * u2pv2);
  if(c_speed < 1e-6) c_speed = 1e-6;    // avoid division by 0 if critical
  c_speed = sqrt(c_speed);                    
  real c_speed2 = c_speed * norme;
  //
  //.. Compute the eigenvalues of the Jacobian matrix
  real eig_val1 = scal - c_speed2;
  real eig_val2 = scal;
  real eig_val3 = scal + c_speed2;
  //
  //.. Compute the ROE flux 
  //.... In this part many tests upon the eigenvalues
  //.... are done to simplify calculations                    
  //.... Here we use the two formes of the ROE flux :
  //.... phi(Wl,Wr) = F(Wl) + A-(Wroe)(Wr - Wl)
  //.... phi(Wl,Wr) = F(Wr) - A+(Wroe)(Wr - Wl)
  //
  if(eig_val2 <= 0.0) {
    p_t = gam1 * (rhoE_t - 0.5 * (rhou_t * rhou_t + 
					  rhov_t * rhov_t) / rho_t);
    ep_t = rhoE_t + p_t;
    rhouv_t = rhou_t * v_t;
    flux_rho  = rhou_t * nx + rhov_t * ny;
    flux_rhou = (rhou_t * u_t + p_t) * nx + rhouv_t * ny;
    flux_rhov = rhouv_t * nx + (rhov_t * v_t + p_t) * ny;
    flux_rhoE = ep_t * (u_t * nx + v_t * ny);
    //
    //.... A Entropic modification
    //
    p_f = gam1 * (rhoE_f - 0.5 * (rhou_f * rhou_f + rhov_f * rhov_f) / rho_f);        
    lambda_f = u_f * nx + v_f * ny + norme 
      * sqrt(gam * p_f / rho_f);
    lambda_t   = u_t * nx + v_t * ny + norme
      * sqrt(gam * p_t / rho_t);
    if ((lambda_f < 0.) && (lambda_t > 0.)) {
      eig_val3 = lambda_t * (eig_val3 - lambda_f) / (lambda_t - lambda_f);
    };
    //           
    if (eig_val3 > 0.0) {
      //.. In this case A+ is obtained by multiplying the last
      //.. colomne of T-1 with the last row of T with eig_val3
      delta_rho  = rho_t - rho_f;
      delta_rhou = rhou_t - rhou_f;
      delta_rhov = rhov_t - rhov_f;
      delta_rhoE = rhoE_t - rhoE_f;
      //
      scal = scal / norme;
      hnx = nx / norme;
      hny = ny / norme;
      usc = 1.0/c_speed;
      tempo11 = gam1 * usc;
      //.. Last columne of the matrix T-1
      a1 = usc;
      a2 = u_ave * usc + hnx;
      a3 = v_ave * usc + hny;
      a4 = 0.5 * u2pv2 * usc + 2.5 * c_speed + scal;
      //.. Last row of the matrix T * eig_val3
      b1 = 0.5 * eig_val3 * (0.5 * tempo11 * u2pv2 - scal);
      b2 = 0.5 * eig_val3 * (hnx - tempo11 * u_ave);
      b3 = 0.5 * eig_val3 * (hny - tempo11 * v_ave);
      b4 = 0.5 * eig_val3 * tempo11;
      //
      alpha1 = a1 * b1 * delta_rho;
      alpha2 = a1 * b2 * delta_rhou;
      alpha3 = a1 * b3 * delta_rhov;
      alpha4 = a1 * b4 * delta_rhoE; 
      alpha = alpha1 + alpha2 + alpha3 + alpha4;
      //
      flux_rho  -= alpha;
      flux_rhou -= a2 * b1 * delta_rho + a2 * b2 * delta_rhou +
	a2 * b3 * delta_rhov + a2 * b4 * delta_rhoE;
      flux_rhov -= a3 * b1 * delta_rho + a3 * b2 * delta_rhou +
	a3 * b3 * delta_rhov + a3 * b4 * delta_rhoE;                         
      flux_rhoE -= a4 * b1 * delta_rho + a4 * b2 * delta_rhou +
	a4 * b3 * delta_rhov + a4 * b4 * delta_rhoE;        
    };
  };
  //
  if(eig_val2 > 0.0) {
    p_f = gam1 * (rhoE_f - 0.5 * (rhou_f * rhou_f +
					  + rhov_f * rhov_f) / rho_f);
    ep_f = rhoE_f + p_f;
    rhouv_f = rhou_f * v_f;
    flux_rho  = rhou_f * nx + rhov_f * ny;
    flux_rhou = (rhou_f * u_f + p_f) * nx + rhouv_f * ny;
    flux_rhov = rhouv_f * nx + (rhov_f * v_f + p_f) * ny;
    flux_rhoE = ep_f * (u_f * nx + v_f * ny);
    //
    // A Entropic modification
    //
    p_t = gam1 * (rhoE_t - 0.5 * (rhou_t * rhou_t +
				      + rhov_t * rhov_t) / rho_t);        
    lambda_f = u_f * nx + v_f * ny - norme 
      * sqrt(gam * p_f / rho_f);
    lambda_t   = u_t * nx + v_t * ny - norme
      * sqrt(gam * p_t / rho_t);
    if ((lambda_f < 0.) && (lambda_t > 0.)) {
      eig_val1 = lambda_f * (lambda_t - eig_val1) / (lambda_t - lambda_f);
    };
    //           
    if (eig_val1 < 0.0) {
      //.. In this case A+ is obtained by multiplying the first
      //.. columne of T-1 with the first row of T with eig_val1
      delta_rho  = rho_t - rho_f;
      delta_rhou = rhou_t - rhou_f;
      delta_rhov = rhov_t - rhov_f;
      delta_rhoE = rhoE_t - rhoE_f;
      //
      scal = scal / norme;        
      hnx = nx / norme;
      hny = ny / norme;
      usc = 1.0/c_speed;
      tempo11 = gam1 * usc;
      //.. First colomne of the matrix T-1
      a1 = usc;
      a2 = u_ave * usc - hnx;
      a3 = v_ave * usc - hny;
      a4 = 0.5 * u2pv2 * usc + 2.5 * c_speed - scal;
      //.. First row of the matrix T * eig_val1
      b1 = 0.5 * eig_val1 * (0.5 * tempo11 * u2pv2 + scal);
      b2 = -0.5 * eig_val1 * (hnx + tempo11 * u_ave);
      b3 = -0.5 * eig_val1 * (hny + tempo11 * v_ave);
      b4 = 0.5 * eig_val1 * tempo11;
      //
      alpha1 = a1 * b1 * delta_rho;
      alpha2 = a1 * b2 * delta_rhou;
      alpha3 = a1 * b3 * delta_rhov;
      alpha4 = a1 * b4 * delta_rhoE; 
      alpha = alpha1 + alpha2 + alpha3 + alpha4;
      //
      flux_rho  += alpha;
      flux_rhou += a2 * b1 * delta_rho + a2 * b2 * delta_rhou +
	a2 * b3 * delta_rhov + a2 * b4 * delta_rhoE;
      flux_rhov += a3 * b1 * delta_rho + a3 * b2 * delta_rhou +
	a3 * b3 * delta_rhov + a3 * b4 * delta_rhoE;                         
      flux_rhoE += a4 * b1 * delta_rho + a4 * b2 * delta_rhou +
	a4 * b3 * delta_rhov + a4 * b4 * delta_rhoE;
    };
  };        
};
  
inline void Roe2D::operate(size_t ie, size_t ip, size_t from, size_t to, 
			   real nx, real ny, bool mod_from, bool mod_to)
{
  real flux_rho, flux_rhou, flux_rhov, flux_rhoE;
  compute(rho_from[ie], rhou_from[ie], rhov_from[ie], rhoE_from[ie],
	  rho_to[ie], rhou_to[ie], rhov_to[ie], rhoE_to[ie],
	  nx, ny, gamma, 
	  flux_rho, flux_rhou, flux_rhov, flux_rhoE);

  //.. Add to pointwise residual
  //.... .._to-side variables
  if (mod_to) {
    res_rho[to] += flux_rho;
    res_rhou[to] += flux_rhou;
    res_rhov[to] += flux_rhov;
    res_rhoE[to] += flux_rhoE;
  };
  //.... .._from-side variables
  if (mod_from) {
    res_rho[from] -= flux_rho;
    res_rhou[from] -= flux_rhou;
    res_rhov[from] -= flux_rhov;
    res_rhoE[from] -= flux_rhoE;
  };
};

#endif



