// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik
// Universitaet Duisburg, Germany
// Institute for Combustion and Gas Dynamics
// University of Duisburg, Germany
// Thu Aug 5 1999
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!
#ifndef MultiGridLeastSquareProlongation_HH
#define MultiGridLeastSquareProlongation_HH

#include <mouse_config.h>

template <class POP> class MultiGridLeastSquareProlongation;

#include <PopList2D.hh>
#include <mouse_types.hh>
#include <ConstructArg.hh>
#include <MultiGridTransferOperator.hh>
#include <geometry2D.hh>

/**
   interpolation from coarse to fine using a least square approach.
   This prolongation operator is very expensive in terms of CPU usage.
*/
template <class POP>
class MultiGridLeastSquareProlongation : public MultiGridTransferOperator<POP>
{
protected:
  /** */ POP *POPL;
  /** */ real M11;
  /** */ real M12;
  /** */ real M13;
  /** */ real M22;
  /** */ real M23;
  /** */ real M33;
  /** */ real M11_inv;
  /** */ real M12_inv;
  /** */ real M13_inv;
  /** */ real M22_inv;
  /** */ real M23_inv;
  /** */ real M33_inv;
  /** */ real *RHS1;
  /** */ real *RHS2;
  /** */ real *RHS3;
  /** */ real *A;
  /** */ real *B;
  /** */ real *C;
  /** */ real min_dist;
  /** */ size_t min_agglo;

public:
  /** */ MultiGridLeastSquareProlongation (carg_t arg);
  /** */ virtual ~MultiGridLeastSquareProlongation ();
  /** */ void contribute(size_t i_fine, size_t i_agglo);
  /** */ void invert();
  /** */ void operate(size_t i_tmp, size_t i_fine);
};

template<class POP>
inline void MultiGridLeastSquareProlongation<POP>::contribute
(size_t i_fine, size_t i_agglo)
{
  // compute weighting factor
  real W;
  if (i_agglo == min_agglo) {
    W = 1;
  } else {
    real dist = DIST(PL_coarse->x(i_agglo), PL_coarse->y(i_agglo),
		     PL->x(i_fine), PL->y(i_fine));
    W = min_dist/dist;
  };
  
  // add contribution of this agglomerate to M
  M11 += 2*W;
  M12 += 2*W * PL_coarse->x(i_agglo);
  M13 += 2*W * PL_coarse->y(i_agglo);
  M22 += 2*W * PL_coarse->x(i_agglo) * PL_coarse->x(i_agglo);
  M23 += 2*W * PL_coarse->x(i_agglo) * PL_coarse->y(i_agglo);
  M33 += 2*W * PL_coarse->y(i_agglo) * PL_coarse->y(i_agglo);
  
  // add contribution of this agglomerate to RHS
  for (size_t iv = 0; iv < VL->NumVars(); iv++) {
    RHS1[iv] += 2*W * var_coarse[iv][i_agglo];
    RHS2[iv] += 2*W * var_coarse[iv][i_agglo] * PL_coarse->x(i_agglo);
    RHS3[iv] += 2*W * var_coarse[iv][i_agglo] * PL_coarse->y(i_agglo);
  };
};

template<class POP>
inline void MultiGridLeastSquareProlongation<POP>::invert()
{
  real a = M11*M22;
  real b = M11*M23;
  real c = M12*M12;
  real d = M13*M12;
  real e = M13*M13;
  real f = 1/(a*M33-b*M23-c*M33+d*M23+d*M23-e*M22);
  M11_inv = (M22*M33-M23*M23)*f;
  M12_inv = -(M12*M33-M13*M23)*f;
  M13_inv = (M12*M23-M13*M22)*f;
  M22_inv = (M11*M33-e)*f;
  M23_inv = -(b-d)*f;
  M33_inv = (a-c)*f;
};

template<class POP>
inline void MultiGridLeastSquareProlongation<POP>::operate
(size_t i_tmp, size_t i_fine)
{
  if (TransferPoint(i_fine)) {

    // get the coarse grid cell of this one
    size_t i_coarse = mg_agent->CoarsePoint(i_fine);

    // compute the minimal distance min_dist and 
    // store the corresponding agglomerate in min_agglo
    min_dist = DIST(PL_coarse->x(i_coarse), PL_coarse->y(i_coarse),
		    PL->x(i_fine), PL->y(i_fine));
    min_agglo = i_coarse;
    for (size_t i = 0; i < POPL->NumFLinks(i_coarse); i++) {
      size_t neigh_agglo = POPL->FLinkPoint(i_coarse, i);
      real dist = DIST(PL_coarse->x(neigh_agglo), PL_coarse->y(neigh_agglo),
		       PL->x(i_fine), PL->y(i_fine));
      if (dist < min_dist) {
	min_dist = dist;
	min_agglo = neigh_agglo;
      };
    };    

    // initialize symmetric matrix M
    M11 = 0; M12 = 0; M13 = 0;
    M22 = 0; M23 = 0;
    M33 = 0;
		      
    //   initialize right hand side vector RHS
    RHS1 = new real[VL->NumVars()];
    RHS2 = new real[VL->NumVars()];
    RHS3 = new real[VL->NumVars()];
    for (size_t iv = 0; iv < VL->NumVars(); iv++) {
      RHS1[iv] = 0;
      RHS2[iv] = 0;
      RHS3[iv] = 0;
    };

    // contribution of this agglomerate
    contribute(i_fine, i_coarse);

    // contributions of all coarse neighbour cells
    for (size_t i = 0; i < POPL->NumFLinks(i_coarse); i++) {
      size_t neigh_agglo = POPL->FLinkPoint(i_coarse, i);
      contribute(i_fine, neigh_agglo);
    };

    // invert M
    invert();
    
    // initialize coefficients of linear function A+Bx+Cy
    A = new real[VL->NumVars()];
    B = new real[VL->NumVars()];
    C = new real[VL->NumVars()];

    // compute coefficients
    for (size_t iv = 0; iv < VL->NumVars(); iv++) {
      A[iv] = M11_inv*RHS1[iv] + M12_inv*RHS2[iv] + M13_inv*RHS3[iv];
      B[iv] = M12_inv*RHS1[iv] + M22_inv*RHS2[iv] + M23_inv*RHS3[iv];
      C[iv] = M13_inv*RHS1[iv] + M23_inv*RHS2[iv] + M33_inv*RHS3[iv];
    };

    // compute correction values on fine grid using A+Bx+Cy
    real corr;
    for (size_t iv = 0; iv < VL->NumVars(); iv++) {
      corr = A[iv] + B[iv]*PL->x(i_fine) + C[iv]*PL->y(i_fine);
      var_fine[iv][i_fine] += factor * corr;
    };

    // delete temporary fields
    delete [] RHS1;
    delete [] RHS2;
    delete [] RHS3;
    delete [] A;
    delete [] B;
    delete [] C;
  };
};

#include "MultiGridLeastSquareProlongation.cc"

#endif








