// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik
// Universitaet Duisburg, Germany
// Institute for Combustion and Gas Dynamics
// University of Duisburg, Germany
// Mon Jun 7 1999
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "SmallSquareMatrix.hh"

#ifndef SmallSquareMatrix_CC
#define SmallSquareMatrix_CC

template <class T, uint N>
SmallSquareMatrix<T, N> SmallSquareMatrix<T, N>::Identity()
{
  SmallSquareMatrix<T, N> I;
  for (uint i = 0; i < N; i++) {
    for (uint j = 0; j < N; j++) {
      if (i==j) I[i][j] = 1;
      else I[i][j] = 0;
    };
  };
  return I;
};

template <class T, uint N>
T SmallSquareMatrix<T,N>::AbsMax()
{
  T max = 0;
  for (uint i = 0; i < N; i++) {
    for (uint j = 0; j < N; j++) {
      if (fabs(V[i][j]) > max) max = fabs(V[i][j]);
    };
  };  
  return max;
};

template <class T, uint N>
void SmallSquareMatrix<T,N>::SwapRows(uint r1, uint r2)
{
  T h;
  for (uint i = 0; i < N; i++) {
    h = V[r1][i];
    V[r1][i] = V[r2][i];
    V[r2][i] = h;
  };
}

template <class T, uint N>
uint SmallSquareMatrix<T,N>::FindNonZero(uint c)
{
  for (uint i = 0; i < N; i++) if (fabs(V[i][c]) > limit) return i;
  return N;
};

template <class T, uint N>
uint SmallSquareMatrix<T,N>::NumLowerZeros(uint r)
{
  uint i = 0;
  uint num = 0;
  while ((i < N) && (fabs(V[r][i]) < limit)) {
    num++;
    i++;
  };
  return num;
};

template <class T, uint N>
uint SmallSquareMatrix<T,N>::NumUpperZeros(uint r)
{
  uint i = N-1;
  uint num = 0;
  while ((i > 0) && (fabs(V[r][i]) < limit)) {
    num++;
    i--;
  };
  return num;
};

template <class T, uint N>
void SmallSquareMatrix<T,N>::Sort(SmallSquareMatrix<T,N> &M)
{
  for (uint i = 0; i < N-1; i++) {
    for (uint j = i; j < N-1; j++) {
      uint lower1 = NumLowerZeros(j);
      uint upper1 = NumUpperZeros(j);
      uint lower2 = NumLowerZeros(j+1);
      uint upper2 = NumUpperZeros(j+1);
      if (lower1 > lower2) {
	SwapRows(j,j+1);
	M.SwapRows(j,j+1);
      } else if (upper1 < upper2) {
	SwapRows(j,j+1);
	M.SwapRows(j,j+1);
      };
    };
  };
  /*
  for (uint i = 0; i < N-1; i++) {
    for (uint j = i+1; j < N; j++) {
      if ((fabs(V[i][i]) < limit) && (fabs(V[j][i]) > limit)) {
	SwapRows(i,j);
	M.SwapRows(i,j);
      };
    };
  };
  */
};

template <class T, uint N>
void SmallSquareMatrix<T,N>::Conditionate(SmallSquareMatrix<T,N> &M)
{
  for (uint i = 0; i < N; i++) {
    T max = 0;
    for (uint j = 0; j < N; j++) {
      if (fabs(V[i][j]) > max) max = fabs(V[i][j]);
      if (fabs(M[i][j]) > max) max = fabs(M[i][j]);
    };
    max = 1.0/max;
    for (uint j = 0; j < N; j++) {
      V[i][j] *= max;
      M[i][j] *= max;
    };
  };
};

template <class T, uint N>
SmallSquareMatrix<T,N> SmallSquareMatrix<T,N>::Inv()
{
  SmallSquareMatrix<T,N> M_left = *this;
  SmallSquareMatrix<T,N> M_right = Identity();
  T maxi;
  maxi = 1.0/M_left.AbsMax();
  for (uint j = 0; j < N; j++) {
    M_left.MultRow(j, maxi);
    M_right.MultRow(j, maxi);
  };
  //M_left.Sort(M_right);
  /*
  for (uint i = 0; i < N; i++) {
    uint j = FindNonZero(i);
    if (i != j) {
      M_left.SwapRows(i,j);
      M_right.SwapRows(i,j);
    };
  };
  */
  M_left.Conditionate(M_right);
//   cout << "left :" << M_left << endl;
//   cout << "right:" << M_right << endl;
//   cout << endl << endl;
  for (uint i = 0; i < N-1; i++) {
    for (uint j = N-1; j > i; j--) {
//       cout << j << ',' << i << endl;
      if (fabs(M_left[j][i]) > limit) {
	if (fabs(M_left[j-1][i]) > limit) {
	  T a = M_left[j][i];
	  T b = M_left[j-1][i];
	  M_left.MultRow(j-1,a);
	  M_right.MultRow(j-1,a);
	  M_left.MultRow(j,b);
	  M_right.MultRow(j,b);
	  M_left.SubstractRow(j-1, j);
	  M_right.SubstractRow(j-1, j);
	} else {
	  M_left.SwapRows(j,j-1);
	  M_right.SwapRows(j,j-1);
	};
      };
      maxi = 1.0/M_left.AbsMax();
      for (uint j = 0; j < N; j++) {
	M_left.MultRow(j, maxi);
	M_right.MultRow(j, maxi);
      };
      //M_left.Sort(M_right);
//       cout << "left :" << M_left << endl;
//       cout << "right:" << M_right << endl;
      M_left.Conditionate(M_right);
//       cout << "left :" << M_left << endl;
//       cout << "right:" << M_right << endl;
//       cout << endl;
    };
  };
  cout << "===============================================" << endl;
  for (uint i = N-1; i > 0; i--) {
    for (uint j = 0; j < i; j++) {
//       cout << j << ',' << i << endl;
      if (fabs(M_left[j][i]) > limit) {
	if (fabs(M_left[j+1][i]) > limit) {
	  T a = M_left[j][i];
	  T b = M_left[j+1][i];
	  M_left.MultRow(j+1,a);
	  M_right.MultRow(j+1,a);
	  M_left.MultRow(j,b);
	  M_right.MultRow(j,b);
	  M_left.SubstractRow(j+1, j);
	  M_right.SubstractRow(j+1, j);
	} else {
	  M_left.SwapRows(j,j+1);
	  M_right.SwapRows(j,j+1);
	};
      };
      maxi = 1.0/M_left.AbsMax();
      for (uint j = 0; j < N; j++) {
	M_left.MultRow(j, maxi);
	M_right.MultRow(j, maxi);
      };
      //M_left.Sort(M_right);
      M_left.Conditionate(M_right);
//       cout << "left :" << M_left << endl;
//       cout << "right:" << M_right << endl;
//       cout << endl;
    };
  };
  for (uint i = 0; i < N; i++) {
    M_right.MultRow(i,1/M_left[i][i]);
  };
  return M_right;
};

#endif

